from itertools import chain

from django.db import models
from django.db.models import Func, fields
from django.contrib.auth import get_user_model
from .constants import CommonStatus
from .fields import (
    AutoUUIDField,
    UpdatedAtField,
    CreatedAtField,
    StatusField,
    DefaultCodeField,
    AuditStatusField,
    CreatedByField, UpdatedByField,
)

User = get_user_model()


class IsNull(Func):
    template = '%(expressions)s IS NULL'
    output_field = fields.BooleanField()
    arity = 1


class NotNull(Func):
    template = '%(expressions)s IS NOT NULL'
    output_field = fields.BooleanField()
    arity = 1


class StatusQueryset(models.QuerySet):
    def editable(self):
        return self.exclude(status__in=[CommonStatus.DELETED, CommonStatus.INVALID])

    def valid(self):
        return self.filter(status__in=[CommonStatus.VALID, CommonStatus.PAUSED, CommonStatus.TO_INVALID])


class BaseModel(models.Model):
    """
    标准抽象模型模型,可直接继承使用
    """

    status = StatusField()  # 状态
    updated_at = UpdatedAtField()  # 修改时间
    created_at = CreatedAtField()  # 创建时间

    objects = models.Manager()

    class Meta:
        abstract = True
        verbose_name = '基本模型'


class BaseCodeModel(models.Model):
    """
    标准抽象模型模型(增加code),可直接继承使用
    """

    code = DefaultCodeField()  # 编号
    status = StatusField()  # 状态
    updated_at = UpdatedAtField()  # 修改时间
    created_at = CreatedAtField()  # 创建时间

    objects = models.Manager()

    class Meta:
        abstract = True
        verbose_name = '基本模型(code)'


class BaseCreatorModel(models.Model):
    """
    审计抽象模型模型,可直接继承使用
    覆盖字段时, 字段名称请勿修改, 必须统一审计字段名称
    """

    status = StatusField()  # 状态
    created_by = CreatedByField()  # 创建者
    updated_by = UpdatedByField()  # 修改者
    updated_at = UpdatedAtField()  # 修改时间
    created_at = CreatedAtField()  # 创建时间

    class Meta:
        abstract = True


class UUIDModel(BaseModel):
    """
    标准抽象模型模型,可直接继承使用
    """

    id = AutoUUIDField()
    updated_at = UpdatedAtField()  # 修改时间
    created_at = CreatedAtField()  # 创建时间

    class Meta:
        abstract = True
        verbose_name = 'UUID模型'


class AuditModel(models.Model):
    """
    审计抽象模型模型,可直接继承使用
    覆盖字段时, 字段名称请勿修改, 必须统一审计字段名称
    """

    status = StatusField()  # 状态
    audit_status = AuditStatusField()  # 审核状态
    created_by = CreatedByField()  # 创建者
    updated_by = UpdatedByField()  # 修改者
    updated_at = UpdatedAtField()  # 修改时间
    created_at = CreatedAtField()  # 创建时间

    class Meta:
        abstract = True
        verbose_name = '审核模型'


class ToDictModelMixin:  # noqa
    def to_dict(self, fields=None, exclude=None, convert_choice=False, fields_map=None):
        """
        Return a dict containing the data in ``instance`` suitable for passing as
        a Form's ``initial`` keyword argument.

        ``fields`` is an optional list of field names. If provided, return only the
        named.

        ``exclude`` is an optional list of field names. If provided, exclude the
        named from the returned dict, even if they are listed in the ``fields``
        argument.

        ``translate_choice`` If provided, convert the value into display value.

        ``field_map`` is dict object, If provided, perform field name mapping.
        """
        opts = self._meta  # noqa
        fields_map = fields_map or {}
        data = {}
        assert not all([fields, exclude]), "Cannot set both 'fields' and 'exclude' options."
        for f in chain(opts.concrete_fields, opts.private_fields):
            if fields and f.name not in fields:
                continue
            if exclude and f.name in fields:
                continue

            field_name = fields_map.get(f.name, f.name)
            if convert_choice and f.choices:
                data[field_name] = getattr(self, f'get_{f.name}_display')()
            else:
                data[field_name] = f.value_from_object(self)

        for f in opts.many_to_many:
            field_name = fields_map.get(f.name, f.name)
            data[field_name] = [i.id for i in f.value_from_object(self)]
        return data



import collections

from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers
from rest_framework.relations import RelatedField
from rest_framework.fields import Field, ChoiceField, to_choices_dict, flatten_choices_dict

__all__ = (
    "SequenceField",
    "DisplayChoiceField",
    "MultiSlugRelatedField",
    "IsNullField",
    "IsNotNullField",
)


class SequenceField(Field):
    """
    A read-only field that output increasing number started from zero.
    """

    cached_attr_name = '_curval'

    def __init__(self, start=1, step=1, **kwargs):
        self.start = start
        self.step = step
        kwargs['source'] = '*'
        kwargs['read_only'] = True
        super().__init__(**kwargs)

    def _incr(self):
        curval = getattr(self, self.cached_attr_name, self.start)
        setattr(self, self.cached_attr_name, curval + self.step)
        return curval

    def to_representation(self, value):
        return self._incr()


class DisplayChoiceField(ChoiceField):
    """
    Serialize: convert value into choice strings
    Deserialize: convert choice strings into value
    """
    def to_representation(self, value):
        if value in ('', None):
            return value
        return self.value_to_display_strings.get(str(value), value)

    def _get_choices(self):
        return self._choices

    def _set_choices(self, choices):
        self.grouped_choices = to_choices_dict(choices)
        self._choices = flatten_choices_dict(self.grouped_choices)
        # Map the string representation of choices to the underlying value.
        # Allows us to deal with eg. integer choices while supporting either
        # integer or string input, but still get the correct datatype out.
        self.value_to_display_strings = {
            str(key): value for key, value in self.choices.items()
        }
        self.display_strings_to_value = {
            str(value): key for key, value in self.choices.items()
        }

    choices = property(_get_choices, _set_choices)


class MultiSlugRelatedField(RelatedField):
    """
    Represents a relationship using a unique set of fields on the target.
    """

    default_error_messages = {
        'does_not_exist': _("Object with {error_msg} does not exist."),
        'invalid': _('Invalid value.'),
    }

    def __init__(self, slug_fields=None, **kwargs):
        assert slug_fields is not None, 'The `slug_fields` argument is required.'
        self.slug_fields = slug_fields
        super().__init__(**kwargs)

    def to_internal_value(self, data):
        if not isinstance(data, collections.Mapping):
            self.fail('invalid')
        if not set(data.keys()) == set(self.slug_fields):
            self.fail('invalid')
        try:
            instance = self.get_queryset().get(**data)
            return instance
        except ObjectDoesNotExist:
            lookups = ['='.join((lookup, value)) for lookup, value in data.items()]
            self.fail('does_not_exist', error_msg=' '.join(lookups))
        except (TypeError, ValueError):
            self.fail('invalid')

    def to_representation(self, value):
        return dict(zip(self.slug_fields, (getattr(value, slug_field) for slug_field in self.slug_fields)))


class StringListField(serializers.ListField):
    child = serializers.CharField()


class IntegerListField(serializers.ListField):
    child = serializers.IntegerField()


class IsNullField(serializers.ReadOnlyField):
    default_error_messages = {
        'invalid': _('Must be a valid boolean.')
    }

    def to_representation(self, value):
        return value is None


class IsNotNullField(IsNullField):

    def to_representation(self, value):
        return value is not None



from django import forms
from django_filters.constants import EMPTY_VALUES
from django_filters.fields import RangeField
from rest_framework.filters import BaseFilterBackend
from django.db.models import Q
from django_filters.filters import (
    CharFilter,
    MultipleChoiceFilter,
    ModelMultipleChoiceFilter,
    DateFromToRangeFilter,
    Filter,
)

import operator
from functools import reduce

from .fields import MultipleValueField, DisplayMultipleChoiceField
from .widgets import ExtendedRangeWidget, ExtendedDateRangeWidget, FixedQueryArrayWidget, LookupTextInput


class SearchFilter(CharFilter):
    """
    This filter performs OR(by default) query on the multi fields.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("lookup_expr", "icontains")
        self.search_fields = kwargs.pop('search_fields', None)
        super().__init__(*args, **kwargs)

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs
        if self.distinct:
            qs = qs.distinct()

        if self.search_fields:
            queries = (Q(**{'%s__%s' % (search_field, self.lookup_expr): value}) for search_field in self.search_fields)
            conditions = reduce(operator.or_, queries)
            qs = qs.filter(conditions)
        else:
            lookup = '%s__%s' % (self.field_name, self.lookup_expr)
            qs = self.filter(**{lookup: value})

        return qs


class MultipleValueFilter(Filter):
    """
    支持传入多个值查询一个字段
    使用示例：stage_name = MultipleValueFilter(field_class=CharField, field_name="stage__name", lookup_expr="icontains")
    支持传参方式：
        1. ?stage_name[]=123&stage_name[]=124
        2. ?stage_name=123&stage_name=125
        3. ?stage_name=123,125
    """
    field_class = MultipleValueField

    def __init__(self, *args, field_class, **kwargs):
        kwargs.setdefault('lookup_expr', 'exact')
        kwargs.setdefault("widget", FixedQueryArrayWidget)
        super().__init__(*args, field_class=field_class, **kwargs)

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs
        if self.distinct:
            qs = qs.distinct()

        lookup = '%s__%s' % (self.field_name, self.lookup_expr)
        queries = (Q(**{lookup: val}) for val in value)
        conditions = reduce(operator.or_, queries)
        qs = self.get_method(qs)(conditions)
        return qs


class DataPermissionFilter(BaseFilterBackend):
    """数据权限filter"""

    @staticmethod
    def data_permission(user, queryset, permission_cls):
        """
        数据权限
        """
        if user.is_anonymous:
            return queryset.none()

        dp = permission_cls(user, queryset)
        queryset = dp.filter()
        return queryset

    def filter_queryset(self, request, queryset, view):
        user = request.user
        queryset = self.data_permission(user, queryset)
        return queryset


class MultiSearchMixin:
    lookup_expr = "icontains"

    def __init__(self, *args, search_fields, **kwargs):
        self.search_fields = search_fields
        kwargs.setdefault("lookup_expr", self.lookup_expr)
        super().__init__(*args, **kwargs)

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        if self.distinct:  # noqa
            qs = qs.distinct()

        queries = (Q(**{'%s__%s' % (search_field, self.lookup_expr): value}) for search_field in self.search_fields)
        conditions = reduce(operator.or_, queries)
        qs = self.get_method(qs)(conditions)  # noqa
        return qs


class MultiSearchFilter(MultiSearchMixin, CharFilter):
    """
    This filter performs OR(by default) query on the multi fields.
    """

    ...


class MultipleChoiceSearchFilter(MultiSearchMixin, ModelMultipleChoiceFilter):
    """
    Extended MultipleChoiceSearchFilter
    """

    lookup_expr = 'in'


class IsNullFilter(Filter):
    field_class = forms.NullBooleanField

    def __init__(self, *args, **kwargs):
        kwargs['lookup_expr'] = 'isnull'
        super().__init__(*args, **kwargs)


class IsNotNullFilter(IsNullFilter):
    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        return super().filter(qs, not value)


class MultipleSelectFilter(Filter):
    def __init__(self, field_name=None, **kwargs):
        kwargs.setdefault("lookup_expr", "in")
        kwargs.setdefault("widget", FixedQueryArrayWidget)
        super().__init__(field_name=field_name, **kwargs)


class ExtendedModelMultipleChoiceFilter(ModelMultipleChoiceFilter):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("widget", FixedQueryArrayWidget)
        super().__init__(*args, **kwargs)


class ExtendedMultipleChoiceFilter(MultipleChoiceFilter):
    always_filter = False

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("widget", FixedQueryArrayWidget)
        super().__init__(*args, **kwargs)

    def is_noop(self, qs, value):
        """
        Return `True` to short-circuit unnecessary and potentially slow
        filtering.
        """
        if self.always_filter:
            return False

        # A reasonable default for being a noop...
        if len(value) == len(self.field.choices):
            return True

        return False


class ExtendedDisplayMultipleChoiceFilter(ExtendedMultipleChoiceFilter):
    field_class = DisplayMultipleChoiceField


class ExtendedRangeFilterMixin:
    lookup_expr = None

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        if value.start is not None and value.stop is not None:
            if value.start == value.stop:
                self.lookup_expr = 'exact'
                value = value.start
            else:
                self.lookup_expr = 'range'
                value = (value.start, value.stop)
        elif value.start is not None:
            self.lookup_expr = 'gte'
            value = value.start
        elif value.stop is not None:
            self.lookup_expr = 'lte'
            value = value.stop

        return super().filter(qs, value)  # noqa


class ExtendedNumberFilter(ExtendedRangeFilterMixin, Filter):
    field_class = RangeField

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("widget", ExtendedRangeWidget)
        super().__init__(*args, **kwargs)


class ExtendedDateFromToRangeFilter(DateFromToRangeFilter):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("widget", ExtendedDateRangeWidget)
        super().__init__(*args, **kwargs)


class ExtendedCharFilter(CharFilter):
    EMPTY_VALUES = EMPTY_VALUES + ("None",)

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("widget", LookupTextInput)
        super().__init__(*args, **kwargs)

    def filter(self, qs, value):
        value, lookup_expr = value.split(":")
        if value in self.EMPTY_VALUES:
            return qs

        self.lookup_expr = lookup_expr
        return super().filter(qs, value)



from collections import OrderedDict
from typing import Dict, List, Optional

from django.utils.encoding import force_str
from rest_framework import serializers
from rest_framework.utils.field_mapping import ClassLookupDict

sentinel = object()

INPUT_TYPE_NONE = None  # 不支持搜索
INPUT_TYPE_INPUT = "input"
INPUT_TYPE_SELECT = "select"
INPUT_TYPE_NUMBER = "num"
INPUT_TYPE_DATE = "date"
INPUT_TYPE_DATETIME = "datetime"
INPUT_TYPE_CHOICE = "select"
INPUT_TYPE_MULTI_CHOICE = "multi_select"


label_lookup = ClassLookupDict(
    {
        serializers.Field: "field",
        serializers.BooleanField: "boolean",
        serializers.NullBooleanField: "boolean",
        serializers.CharField: "string",
        serializers.UUIDField: "string",
        serializers.URLField: "url",
        serializers.EmailField: "email",
        serializers.RegexField: "regex",
        serializers.SlugField: "slug",
        serializers.IntegerField: "integer",
        serializers.FloatField: "float",
        serializers.DecimalField: "decimal",
        serializers.DateField: "date",
        serializers.DateTimeField: "datetime",
        serializers.TimeField: "time",
        serializers.ChoiceField: "choice",
        serializers.MultipleChoiceField: "multiple choice",
        serializers.FileField: "file upload",
        serializers.ImageField: "image upload",
        serializers.ListField: "list",
        serializers.DictField: "nested object",
        serializers.Serializer: "nested object",
    }
)


search_type_lookup = ClassLookupDict(
    {
        serializers.CharField: INPUT_TYPE_INPUT,
        serializers.BooleanField: INPUT_TYPE_CHOICE,
        serializers.NullBooleanField: INPUT_TYPE_CHOICE,
        serializers.IntegerField: INPUT_TYPE_NUMBER,
        serializers.FloatField: INPUT_TYPE_INPUT,
        serializers.DecimalField: INPUT_TYPE_INPUT,
        serializers.DateField: INPUT_TYPE_DATE,
        serializers.DateTimeField: INPUT_TYPE_DATE,
        serializers.TimeField: INPUT_TYPE_DATE,
        serializers.ChoiceField: INPUT_TYPE_CHOICE,
        serializers.MultipleChoiceField: INPUT_TYPE_MULTI_CHOICE,
    }
)

search_key_lookup = ClassLookupDict(
    {
        serializers.Field: "icontains",
        serializers.BooleanField: "",
        serializers.NullBooleanField: "",
        serializers.CharField: "icontains",
        serializers.UUIDField: "icontains",
        serializers.URLField: "icontains",
        serializers.EmailField: "icontains",
        serializers.RegexField: "regex",
        serializers.SlugField: "slug",
        serializers.IntegerField: "",
        serializers.FloatField: "",
        serializers.DecimalField: "",
        serializers.DateField: "range",
        serializers.DateTimeField: "range",
        serializers.TimeField: "time",
        serializers.ChoiceField: "",
        serializers.MultipleChoiceField: "",
    }
)


class DisplayStyle(dict):
    __slots__ = (
        "column_title",
        "column_key",
        "column_align",
        "column_customer_width",
        "column_fixed",
        "column_visible",
        "column_editable",
        "column_sortable",
        "column_search_type",
        "column_search_key",
        "column_search_placeholder",
        "column_show_customize",
        "select_options_type",
        "select_options_value",
        "select_post_url",
        "select_post_data",
        "select_post_key",
        "select_post_name",
        "column_tooltip",
        "group_by_column",
        "child",
        "children",
    )

    def __init__(
        self,
        column_title: str = sentinel,
        column_key: str = sentinel,
        column_align: str = sentinel,
        column_customer_width: int = sentinel,
        column_search_type: str = sentinel,
        column_search_key: str = sentinel,
        column_search_placeholder: str = sentinel,
        column_show_customize: dict = sentinel,
        select_options_type: bool = sentinel,
        select_options_value: str = sentinel,
        select_post_url: str = sentinel,
        select_post_data: str = sentinel,
        select_post_key: str = sentinel,
        select_post_name: str = sentinel,
        column_tooltip: bool = sentinel,
        column_fixed: bool = sentinel,
        column_visible: bool = True,
        column_editable: bool = False,
        column_sortable: bool = False,
        group_by_column: str = sentinel,
        **kwargs
    ):
        self.column_title = column_title
        self.column_key = column_key
        self.column_align = column_align
        self.column_customer_width = column_customer_width
        self.column_fixed = column_fixed
        self.column_visible = column_visible
        self.column_editable = column_editable
        self.column_sortable = column_sortable
        self.column_search_type = column_search_type
        self.column_search_key = column_search_key
        self.column_search_placeholder = column_search_placeholder
        self.column_show_customize = column_show_customize
        self.select_options_type = select_options_type
        self.select_options_value = select_options_value
        self.select_post_url = select_post_url
        self.select_post_data = select_post_data
        self.select_post_key = select_post_key
        self.select_post_name = select_post_name
        self.column_tooltip = column_tooltip
        self.group_by_column = group_by_column
        d = dict()
        for attr in self.__slots__:
            if getattr(self, attr, sentinel) is not sentinel:
                d.update({attr: getattr(self, attr)})
        super().__init__(d, **kwargs)

    @property
    def is_api_select(self):
        return "select_options_type" in self and not self["select_options_type"]

    def set_column_title(self, metacls, field):
        """
        列显示的标题
        """
        self["column_title"] = field.label

    def set_column_key(self, metacls, field):
        """
        列内容的字段key, 即接口返回的数据字段名称
        """
        self["column_key"] = field.field_name

    def set_column_align(self, metacls, field):
        """
        列对齐方式:
          - left
          - right
          - center
        """
        ...

    def set_column_customer_width(self, metacls, field):
        """
        用户自定义列宽，字符串型数字
        """
        ...

    def set_column_fixed(self, metacls, field):
        """
        当前列是否固定,非必填,不填表示不固定
        options: [left, right]
        """
        ...

    def set_column_visible(self, metacls, field):
        """
        当前列是否显示
        """
        ...

    def set_column_editable(self, metacls, field):
        """
        列是否支持编辑
        rtype: bool
        """
        ...

    def set_column_sortable(self, metacls, field):
        """
        列是否支持排序
        """
        if isinstance(field, (serializers.SerializerMethodField, serializers.Serializer)):
            self["column_sortable"] = False
        elif isinstance(field, (serializers.DateField, serializers.DateTimeField)):
            self["column_sortable"] = True

    def set_column_search_type(self, metacls, field):
        """
        列搜索类型，默认空不搜索
        input: 为搜索框，
        select: 下拉选择，
        date: 为时间选择器，
        num: 为数字区间
        """
        if "column_search_type" not in self:
            try:
                if "set_select_post_url" in self:
                    self["column_search_type"] = INPUT_TYPE_SELECT
                else:
                    self["column_search_type"] = search_type_lookup[field]
            except KeyError:
                pass

    def set_column_search_key(self, metacls, field):
        """
        搜索的key值, 非必填
        """
        ...
        # [
        #     (field.source.replace(".", "__") or field_name, field.label)
        #     for field_name, field in serializer_class(context=context).fields.items()
        #     if not getattr(field, "write_only", False) and not field.source == "*"
        # ]

    def set_column_search_placeholder(self, metacls, field):
        """
        输入框提示语, 非必填
        """
        ...

    def set_column_show_customize(self, metacls, field):
        """
        自定义列值，前端处理数据，后端只需要传值
        """
        ...

    def set_select_options_type(self, metacls, field):
        """
        表头下拉选项的值
          - true: 取optionKey的缓冲key值
          - false: 走接口的请求参数
        """
        ...

    def set_select_options_value(self, metacls, field):
        """
        key值, 有两种情况取值不同
        1. 字段提供choices属性
        2. 布尔类型
        """
        if hasattr(field, "choices") and not isinstance(field, serializers.RelatedField):
            self["select_options_value"] = [
                dict(name=force_str(name, strings_only=True), value=value) for value, name in field.choices.items()
            ]
        elif isinstance(field, (serializers.BooleanField, serializers.NullBooleanField)):
            self["select_options_value"] = [{"name": "true", "value": True}, {"name": "false", "value": False}]

    def set_select_post_url(self, metacls, field):
        """
        下拉接口的请求接口
        select_options_type为false时必传
        """
        ...

    def set_select_post_data(self, metacls, field):
        """
        下拉接口的请求接口参数参数
        """
        if self.is_api_select and "select_post_name" not in self:
            self["select_post_key"] = "name"

    def set_select_post_key(self, metacls, field):
        """
        下拉接口的请求key的对应的数据字段名
        """
        ...

    def set_select_post_name(self, metacls, field):
        """
        下拉接口的请求name的对应的数据字段名
        """
        if self.is_api_select and "select_post_name" not in self:
            self["select_post_name"] = "name"

    def set_column_tooltip(self, metacls, field):
        """
        文本溢出隐藏后是否显示提示框
        """
        ...



import operator
from functools import reduce

from django.db.models import Q
from django.forms import MultipleChoiceField


class MultipleValueField(MultipleChoiceField):
    def __init__(self, *args, field_class, **kwargs):
        self.inner_field = field_class()
        super().__init__(*args, **kwargs)

    def valid_value(self, value):
        return self.inner_field.validate(value)

    def clean(self, values):
        return values and [self.inner_field.clean(value) for value in values]


class DisplayMultipleChoiceField(MultipleChoiceField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.choices = [(v, k) for k, v in self.choices]

    def clean(self, value):
        """
        Validate the given value and return its "cleaned" value as an
        appropriate Python object. Raise ValidationError for any errors.
        """
        value = super().clean(value)
        string_to_value_map = dict(self.choices)
        return [string_to_value_map.get(val) for val in value]



import logging
import os
import random
from collections import OrderedDict
from datetime import datetime

from django.core.serializers.json import DjangoJSONEncoder
from django.db.transaction import atomic
from django.utils import timezone


class CustomEncoder(DjangoJSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')

        return super().default(obj)


class MakeFileHandler(logging.FileHandler):
    def __init__(self, filename, mode="a", encoding=None, delay=0):
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        super().__init__(filename, mode, encoding, delay)


@atomic
def atomic_call(*funcs):
    """
    Call function atomicly
    """
    for func in funcs:
        if not callable(func):
            raise TypeError(f"{func} must be callable!")

        func()


def strtobool(val):
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    elif val in ('n', 'no', 'f', 'false', 'off', '0'):
        return False
    else:
        raise ValueError("invalid truth value %r" % (val,))


def get_serial_code(prefix=""):
    """
    生成序列号
    """
    random_number = random.randint(0, 999)
    return timezone.now().strftime(f"{prefix}%y%m%d%H%M%I{random_number:03d}")


def to_table_choices(choices: OrderedDict):
    """
    转换为前端适配的options
    """
    return [{"label": label, "value": value} for value, label in choices.items()]



from django.db import models


class SimpleStatus(models.IntegerChoices):
    VALID = 50, "已生效"
    INVALID = 100, "已失效"


class CommonStatus(models.IntegerChoices):
    DELETED = 0, "已删除"
    TO_SUBMIT = 5, "待提交"
    TO_VALID = 10, "待生效"
    PAUSED = 25, "暂停中"
    VALID = 50, "已生效"
    TO_INVALID = 75, "待失效"
    INVALID = 100, "已失效"


class AuditStatus(models.IntegerChoices):
    TO_AUDIT = 1, "提交审核"
    PASSED = 10, "审核通过"
    REJECTED = 11, "审核拒绝"
    TO_AUDIT_LEVEL2 = 2, "提交审核"
    PASSED_LEVEL2 = 20, "审核通过"
    REJECTED_LEVEL2 = 21, "审核拒绝"
    TO_AUDIT_LEVEL3 = 3, "提交审核"
    PASSED_LEVEL3 = 30, "审核通过"
    REJECTED_LEVEL3 = 31, "审核拒绝"



from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.utils.translation import gettext_lazy as _


class WithoutCountPaginator(Paginator):
    """
    This is a PAGINATOR, NOT PAGINATION
    """

    has_next_page: bool = True
    has_previous_page: bool = False

    def page(self, number):
        number = self.validate_number(number)
        self.has_previous_page = False if number == 1 else True
        bottom = (number - 1) * self.per_page
        top = bottom + self.per_page
        top_with_extra = top + 1
        object_with_extra = list(self.object_list[bottom:top_with_extra])
        if not object_with_extra:
            raise EmptyPage(_('That page contains no results'))
        if len(object_with_extra) >= self.per_page:
            object_with_extra = object_with_extra[:-1]
            self.has_next_page = True
        else:
            self.has_next_page = False
        return self._get_page(object_with_extra, number, self)

    @property
    def count(self):
        return 0

    def validate_number(self, number):
        """Validate the given 1-based page number."""
        try:
            if isinstance(number, float) and not number.is_integer():
                raise ValueError
            number = int(number)
        except (TypeError, ValueError):
            raise PageNotAnInteger(_('That page number is not an integer'))
        if number < 1:
            raise EmptyPage(_('That page number is less than 1'))
        return number






from rest_framework.routers import DefaultRouter


class OptionalSlashRouter(DefaultRouter):
    """
    Providing optional tailing slash for the router.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trailing_slash = '/?'



import functools
import operator
from decimal import Decimal

import orjson
from typing import Optional, Any
import unicodecsv as csv
from io import BytesIO
import datetime

from django.conf import settings
from openpyxl import Workbook
from openpyxl.styles import Font, PatternFill, Alignment

from rest_framework import status
from django.utils.encoding import force_str

from rest_framework.settings import api_settings
from rest_framework.renderers import BaseRenderer
from rest_framework.status import is_success
from django.utils.functional import Promise
from django.db.models.query import QuerySet

__all__ = ["CustomJSONRenderer", "CustomCSVRenderer", "CustomExcelRenderer"]


class CustomJSONRenderer(BaseRenderer):
    """
    Renderer which serializes to JSON.
    Uses the Rust-backed orjson library for serialization speed.
    """

    media_type = "application/json"
    html_media_type = "text/html"
    format = "json"
    charset = None

    options = functools.reduce(
        operator.or_,
        api_settings.user_settings.get("ORJSON_RENDERER_OPTIONS", ()),
        orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_PASSTHROUGH_DATETIME,
    )

    @staticmethod
    def default(obj: Any) -> Any:
        """
        When orjson doesn't recognize an object type for serialization it passes
        that object to this function which then converts the object to its
        native Python equivalent.

        :param obj: Object of any type to be converted.
        :return: native python object
        """

        if isinstance(obj, Promise):
            return force_str(obj)
        elif isinstance(obj, datetime.datetime):
            return obj.strftime(api_settings.DATETIME_FORMAT)
        elif isinstance(obj, Decimal):
            if api_settings.COERCE_DECIMAL_TO_STRING:
                return str(obj)
            else:
                return float(obj)
        elif isinstance(obj, QuerySet):
            return tuple(obj)
        elif hasattr(obj, "tolist"):
            return obj.tolist()
        elif hasattr(obj, "__iter__"):
            return list(item for item in obj)

    def render(
        self,
        data: Any,
        media_type: Optional[str] = None,
        renderer_context: Any = None,
    ) -> bytes:
        """
        Serializes Python objects to JSON.

        :param data: The response data, as set by the Response() instantiation.
        :param media_type: If provided, this is the accepted media type, of the
                `Accept` HTTP header.
        :param renderer_context: If provided, this is a dictionary of contextual
                information provided by the view. By default this will include
                the following keys: view, request, response, args, kwargs
        :return: bytes() representation of the data encoded to UTF-8
        """
        if response := renderer_context.get('response'):
            payload = {}
            if hasattr(renderer_context.get("request"), "id"):
                payload["request_id"] = renderer_context["request"].id

            payload["ret"] = response.status_code
            payload["msg"] = "success"

            if data is not None:
                payload["data"] = data

            if not is_success(response.status_code):
                try:
                    payload["msg"] = data["detail"]
                    payload.pop("data", None)
                except KeyError:
                    payload["msg"] = "Invalid input."

            response.status_code = status.HTTP_200_OK  # Set all response status to HTTP 200
        elif data is None:
            return b""
        else:
            payload = data

        # If `indent` is provided in the context, then pretty print the result.
        # E.g. If we're being called by RestFramework's BrowsableAPIRenderer.
        options = self.options
        if media_type == self.html_media_type:
            options |= orjson.OPT_INDENT_2

        response._rendered_data = payload  # for loging response use
        serialized: bytes = orjson.dumps(payload, default=self.default, option=options)
        return serialized


class BaseExportRenderer(BaseRenderer):
    def validate(self, data: dict):
        return True

    def get_export_data(self, data: dict):
        return data["results"] if "results" in data else data

    def get_file_name(self, renderer_context: Optional[dict]):
        return f'export({datetime.datetime.now().strftime("%Y%m%d")})'

    def tablize(self, data, header=None, labels=None, value_mapping=None):
        """
        Convert a list of data into a table.

        If there is a header provided to tablize it will efficiently yield each
        row as needed. If no header is provided, tablize will need to process
        each row in the data in order to construct a complete header. Thus, if
        you have a lot of data and want to stream it, you should probably
        provide a header to the renderer (using the `header` attribute, or via
        the `renderer_context`).
        """
        # Try to pull the header off of the data, if it's not passed in as an
        # argument.
        if not header and hasattr(data, 'header'):
            header = data.header

        if data:
            # First, flatten the data (i.e., convert it to a list of
            # dictionaries that are each exactly one level deep).  The key for
            # each item designates the name of the column that the item will
            # fall into.
            data = self.flatten_data(data)
            # Get the set of all unique headers, and sort them (unless already provided).
            if not header:
                # We don't have to materialize the data generator unless we
                # have to build a header.
                data = tuple(data)
                header_fields = set()
                for item in data:
                    header_fields.update(list(item.keys()))
                header = sorted(header_fields)

            # Return your "table", with the headers as the first row.
            if labels:
                yield [labels.get(x, x) for x in header]
            else:
                yield header
            # Create a row for each dictionary, filling in columns for which the
            # item has no data with None values.
            for item in data:
                if value_mapping:
                    row = [
                        value_mapping[key].get(item.get(key), item.get(key)) if key in value_mapping else item.get(key)
                        for key in header
                    ]
                else:
                    row = [item.get(key) for key in header]
                yield row
        elif header:
            # If there's no data but a header was supplied, yield the header.
            if labels:
                yield [labels.get(x, x) for x in header]
            else:
                yield header
        else:
            # Generator will yield nothing if there's no data and no header
            pass

    def flatten_data(self, data, value_mapping=None):
        """
        Convert the given data collection to a list of dictionaries that are
        each exactly one level deep. The key for each value in the dictionaries
        designates the name of the column that the value will fall into.
        """
        for item in data:
            yield dict(item)


class CustomCSVRenderer(BaseExportRenderer):
    """
    Renderer which serializes to CSV
    """

    media_type = 'text/csv'
    format = 'csv'
    header = None
    labels = None  # {'<field>':'<label>'}
    writer_opts = None
    data_key = "results"

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """
        Renders serialized *data* into CSV. For a dictionary:
        """
        renderer_context = renderer_context or {}
        if data is None:
            return bytes()

        if isinstance(data, dict):
            data = data[self.data_key]

        writer_opts = renderer_context.get('writer_opts', self.writer_opts or {})
        header = renderer_context.get('header', self.header)
        labels = renderer_context.get('labels', self.labels)
        value_mapping = renderer_context.get('value_mapping')
        encoding = renderer_context.get('encoding', settings.DEFAULT_CHARSET)

        table = self.tablize(data, header=header, labels=labels, value_mapping=value_mapping)
        csv_buffer = BytesIO()
        csv_writer = csv.writer(csv_buffer, encoding=encoding, **writer_opts)
        for row in table:
            csv_writer.writerow(row)

        filename = self.get_file_name(renderer_context)
        renderer_context["response"]['Content-Disposition'] = f'attachment; filename="{filename}.csv"'
        return csv_buffer.getvalue()


class CustomExcelRenderer(BaseExportRenderer):
    """
    Renderer for Excel spreadsheet open data format (xlsx).
    """

    media_type = "application/xlsx"
    format = "xlsx"
    header = None
    labels = None  # {'<field>':'<label>'}
    boolean_labels = None
    custom_mappings = None
    letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    data_key = "results"
    header_font = Font(b=True)
    header_fill = PatternFill('solid', start_color="87CEFA")
    header_height = 17
    freeze_header = True

    def excel_style(self, row, col):
        """Convert given row and column number to an Excel-style cell name."""
        result = []
        while col:
            col, rem = divmod(col - 1, 26)
            result[:0] = self.letters[rem]
        return ''.join(result) + str(row)

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """
        Render `data` into XLSX workbook, returning a workbook.
        """
        if not self.validate(data):
            return bytes()

        if isinstance(data, dict):
            data = data[self.data_key]

        header = renderer_context.get('header', self.header)
        labels = renderer_context.get('labels', self.labels)
        value_mapping = renderer_context.get('value_mapping')

        table = self.tablize(data, header=header, labels=labels, value_mapping=value_mapping)
        excel_buffer = BytesIO()

        workbook = Workbook()
        sheet = workbook.active

        for row in table:
            sheet.append(row)

        for cell in sheet["1:1"]:
            cell.font = self.header_font
            cell.fill = self.header_fill
            cell.alignment = Alignment(vertical='center')

        sheet.row_dimensions[1].height = self.header_height
        if self.freeze_header:
            sheet.freeze_panes = f"A2"

        sheet.print_title_rows = '1:1'
        workbook.save(excel_buffer)

        filename = self.get_file_name(renderer_context)
        renderer_context["response"]['Content-Disposition'] = f'attachment; filename="{filename}.xlsx"'
        return excel_buffer.getvalue()



"""
Global exception and warning classes.
"""
import traceback

from django.conf import settings
from rest_framework.response import Response
from rest_framework.status import HTTP_500_INTERNAL_SERVER_ERROR
from rest_framework.views import set_rollback, exception_handler


def _get_detail(detail):
    if isinstance(detail, list):
        return ''.join([x if isinstance(x, str) else _get_detail(x) for x in detail])
    result = ''
    for v in detail.values():
        result += _get_detail(v)
    return result


def custom_exception_handler(exc, context):
    # Call REST framework's default exception handler first,
    # to get the standard error response.
    response = exception_handler(exc, context)

    # Catch other exceptions.
    if response is None:
        error_message = getattr(exc, "message", repr(exc))
        response = Response({"detail": error_message}, status=HTTP_500_INTERNAL_SERVER_ERROR)
        set_rollback()
        if settings.DEBUG:
            traceback.print_exc()
            response.data.update(traceback=traceback.format_exc())

    return response



from urllib import parse

from django.core.paginator import InvalidPage
from django.utils.encoding import force_str
from rest_framework import pagination
from rest_framework.pagination import CursorPagination
from rest_framework.response import Response

from .paginators import WithoutCountPaginator


class CustomPagination(pagination.PageNumberPagination):
    page_size_query_param = "page_size"
    max_page_size = 5000

    def paginate_queryset(self, queryset, request, view=None):
        page_num = request.query_params.get(self.page_query_param)
        # 判断，如果 page 为all 则取消分页返回所有
        if page_num == 'all':
            request.query_params._mutable = True
            request.query_params[self.page_query_param] = 1
            request.query_params[self.page_size_query_param] = self.max_page_size
            request.query_params._mutable = False

        return super().paginate_queryset(queryset, request, view)

    def get_paginated_response(self, data):
        return Response(
            {
                'total': self.page.paginator.count,
                'page_size': self.page.paginator.per_page,
                'current_page': self.page.number,
                'results': data,
            }
        )

    def get_paginated_response_schema(self, schema):
        return {
            'type': 'object',
            'properties': {
                'total': {
                    'type': 'integer',
                    'example': 123,
                },
                'page_size': {
                    'type': 'integer',
                    'example': 15,
                },
                'current_page': {
                    'type': 'integer',
                    'example': 1,
                },
                'results': schema,
            },
        }


class WithoutCountPagination(CustomPagination):
    has_next: bool = True

    def paginate_queryset(self, queryset, request, view=None):
        """
        Paginate a queryset if required, either returning a
        page object, or `None` if pagination is not configured for this view.
        """
        page_size = self.get_page_size(request)
        if not page_size:
            return None

        paginator = WithoutCountPaginator(queryset, page_size)
        page_number = request.query_params.get(self.page_query_param, 1)
        if page_number in self.last_page_strings:
            page_number = paginator.num_pages
        self.request = request

        try:
            self.page = paginator.page(page_number)
        except InvalidPage:
            return []

        return self.page

    def get_next_link(self):
        if not getattr(self, "page", 0) or not self.page.paginator.has_next_page:
            return None
        url = self.request.build_absolute_uri()
        page_number = self.page.next_page_number()
        return self.replace_query_param(url, self.page_query_param, page_number)

    def get_previous_link(self):
        if not getattr(self, "page", 0) or not self.page.paginator.has_previous_page:
            return None
        url = self.request.build_absolute_uri()
        page_number = self.page.previous_page_number()
        if page_number == 1:
            return self.remove_query_param(url, self.page_query_param)
        return self.replace_query_param(url, self.page_query_param, page_number)

    @staticmethod
    def replace_query_param(url, key, val):
        """
        Given a URL and a key/val pair, set or replace an item in the query
        parameters of the URL, and return the new URL.
        """
        (_, _, path, query, fragment) = parse.urlsplit(force_str(url))
        query_dict = parse.parse_qs(query, keep_blank_values=True)
        query_dict[force_str(key)] = [force_str(val)]
        query = parse.urlencode(sorted(list(query_dict.items())), doseq=True)
        return parse.urlunsplit(("", "", path, query, fragment))

    @staticmethod
    def remove_query_param(url, key):
        """
        Given a URL and a key/val pair, remove an item in the query
        parameters of the URL, and return the new URL.
        """
        (_, _, path, query, fragment) = parse.urlsplit(force_str(url))
        query_dict = parse.parse_qs(query, keep_blank_values=True)
        query_dict.pop(key, None)
        query = parse.urlencode(sorted(list(query_dict.items())), doseq=True)
        return parse.urlunsplit(("", "", path, query, fragment))

    def get_paginated_response(self, data):
        return Response({"list": data, "next": self.get_next_link(), "previous": self.get_previous_link()})


class BigPagePagination(CustomPagination):
    page_size = 5000
    page_size_query_param = "page_size"
    max_page_size = 100000


class CursorSetPagination(CursorPagination):
    page_size = 10
    max_page_size = 500
    page_size_query_param = "page_size"
    ordering = "-created_at"

    def paginate_queryset(self, queryset, request, view=None):
        page = super().paginate_queryset(queryset, request, view=view)
        self.base_url = request.get_full_path()
        return page

    def get_paginated_response(self, data):
        return Response({"previous": self.get_previous_link(), "next": self.get_next_link(), "list": data})



# http://www.django-rest-framework.org/api-guide/settings/
REST_FRAMEWORK = {
    # Use Django's standard `django.contrib.auth` permissions,
    # or allow read-only access for unauthenticated users.
    'DEFAULT_PERMISSION_CLASSES': [
        'permissions.IsAuthenticatedWriteOrReadOnly',
    ],
    'DEFAULT_AUTHENTICATION_CLASSES': [
        # 'oauth2_provider.contrib.rest_framework.OAuth2Authentication',  # Own oauth server
        'client_authentication.ApiTokenAuthentication',
        'rest_framework.authentication.SessionAuthentication',
        'rest_framework.authentication.BasicAuthentication',
    ],
    # Enable DRF pagination
    # 'DEFAULT_PAGINATION_CLASS': 'rest_framework.pagination.PageNumberPagination',
    'DEFAULT_PAGINATION_CLASS': 'drf_defaults.DefaultResultsSetPagination',
    'PAGE_SIZE': 50,
    'DEFAULT_RENDERER_CLASSES': (
        # 'rest_framework.renderers.JSONRenderer',  # Swapping out the original renderer
        'lib.drf_renderer.UJSONRenderer',
        'rest_framework.renderers.BrowsableAPIRenderer',
    ),
    'DEFAULT_PARSER_CLASSES': (
        # 'rest_framework.parsers.JSONParser',  # Swapping out the original parser
        'lib.drf_parser.UJSONParser',
        'rest_framework.parsers.FormParser',
        'rest_framework.parsers.MultiPartParser'
    ),
    'DEFAULT_FILTER_BACKENDS': (
        'django_filters.rest_framework.DjangoFilterBackend',
    )
}


import re

from django.forms import Select, TextInput
from django.utils.datastructures import MultiValueDict
from django_filters.widgets import RangeWidget, QueryArrayWidget


class ExtendedSelectMultiple(Select):
    allow_multiple_selected = True

    def value_from_datadict(self, data, files, name):
        values = list()
        pattern = rf"{name}(\[\d*\])?"
        for k, v in data.items():
            if re.fullmatch(pattern, k):
                values.append(v)
        return values

    def value_omitted_from_data(self, data, files, name):
        # An unselected <select multiple> doesn't appear in POST data, so it's
        # never known if the value is actually omitted.
        return False


class FixedQueryArrayWidget(QueryArrayWidget):
    """
    Enables request query array notation that might be consumed by MultipleChoiceFilter

    1. Values can be provided as csv string:  ?foo=bar,baz
    2. Values can be provided as query array: ?foo[]=bar&foo[]=baz
    3. Values can be provided as query array: ?foo=bar&foo=baz

    Note: Duplicate and empty values are skipped from results
    """

    def value_from_datadict(self, data, files, name):
        if not isinstance(data, MultiValueDict):
            data = MultiValueDict(data)

        values_list = data.getlist(name, data.getlist('%s[]' % name)) or []

        # treat value as csv string: ?foo=1,2
        if len(values_list) == 1:
            ret = [x.strip() for x in values_list[0].rstrip(',').split(',') if x]
        # apparently its an array, so no need to process it's values as csv
        # ?foo=1&foo=2 -> data.getlist(foo) -> foo = [1, 2]
        # ?foo[]=1&foo[]=2 -> data.getlist(foo[]) -> foo = [1, 2]
        elif len(values_list) > 1:
            ret = [x for x in values_list if x]
        else:
            ret = []

        return list(set(ret))


class ExtendedRangeWidget(RangeWidget):
    """
    支持以下格式查询参数
    1. query/company/?a_min=1&a_max=10 -> data=[1,10]
    2. query/company/?a_min=1 -> data=[1, None]
    3. query/company/?a_max=10 -> data=[None, 10]
    4. query/company/?a=1 -> data=[1,1]
    """

    def value_from_datadict(self, data, files, name):
        return [
            widget.value_from_datadict(data, files, self.suffixed(name, suffix))
            or widget.value_from_datadict(data, files, name)
            for widget, suffix in zip(self.widgets, self.suffixes)
        ]


class ExtendedDateRangeWidget(ExtendedRangeWidget):
    suffixes = ['after', 'before']


class LookupTextInput(TextInput):
    suffix = "equal"
    default_lookup_expr = "icontains"

    def suffixed(self, name):
        return f"{name}_{self.suffix}"

    def value_from_datadict(self, data, files, name):
        """
        Given a dictionary of data and this widget's name, return the value
        of this widget or None if it's not provided.
        """
        if self.suffixed(name) in data:
            value = data.get(self.suffixed(name))
            lookup_expr = "exact"
        else:
            value = data.get(name)
            lookup_expr = self.default_lookup_expr

        return f"{value}:{lookup_expr}"



from django.db.models import QuerySet
from rest_framework.relations import ManyRelatedField, RelatedField
from rest_framework.response import Response
from rest_framework.serializers import ModelSerializer, Serializer
from rest_framework.fields import ReadOnlyField
from django.core.exceptions import ImproperlyConfigured
from rest_framework.settings import api_settings
from rest_framework.viewsets import GenericViewSet
from drfexts.metadata import VueTableMetadata
from drfexts.renderers import CustomExcelRenderer, CustomCSVRenderer


class EagerLoadingMixin:
    function_name = "setup_eager_loading"

    def get_queryset(self, *args, **kwargs):
        """
        Call setup_eager_loading function on serializer
        """
        queryset = super().get_queryset(*args, **kwargs)
        serilaizer_class = self.get_serializer_class()  # noqa
        if hasattr(serilaizer_class, "setup_eager_loading") and callable(serilaizer_class.setup_eager_loading):
            queryset = serilaizer_class.setup_eager_loading(queryset)
            assert isinstance(queryset, QuerySet), (
                f"Expected '{self.function_name}' to return a QuerySet, "
                f"but got a {type(queryset).__name__} instead."
            )

        return queryset


class SelectOnlyMixin:
    """
    Mixin used to define select-only fields for queryset
    Cautions:
        1. The mixin is intended for performance optimization and you don't need it in most cases.
    """

    # If using Django filters in the API, these labels mustn't conflict with any model field names.
    include_only_fields_name = "only_fields"
    expand_only_fields_name = "expand_only_fields"
    exclude_only_fields_name = "exclude_only_fields"

    def get_queryset(self):
        """
        Select only fields
        """
        queryset = super().get_queryset()
        serilaizer_class = self.get_serializer_class()  # noqa

        assert issubclass(
            serilaizer_class, ModelSerializer
        ), f'Class {serilaizer_class.__class__.__name__} must inherit from "ModelSerializer"'

        if getattr(queryset, "_result_cache", None):
            return queryset

        meta = getattr(serilaizer_class, "Meta", None)
        only_fields = getattr(meta, self.include_only_fields_name, None)
        expand_only_fields = set(getattr(meta, self.expand_only_fields_name, []))
        # You may need to set this attribute when fetch attrs in `SerializerMethod`
        # or in a nested serializer
        exclude_query_fields = set(getattr(meta, self.exclude_only_fields_name, []))
        if only_fields and exclude_query_fields:
            raise ImproperlyConfigured("You cannot set both 'only_fields' and 'exclude_only_fields'.")

        if only_fields:
            return queryset.only(*only_fields)

        only_fields_name = set()
        for field in serilaizer_class()._readable_fields:
            if field.field_name in exclude_query_fields:
                continue

            # TODO: support nested serializer
            if isinstance(field, (ReadOnlyField, Serializer)):
                continue

            source = getattr(field, "source", None)
            # serliazer method class will set source to '*'
            if source == "*":
                continue

            if source:
                query_name = "__".join(source.split("."))
            else:
                query_name = field.field_name

            only_fields_name.add(query_name)

        only_fields_name |= expand_only_fields
        if only_fields_name:
            queryset = queryset.only(*only_fields_name)

        return queryset


class DynamicListModelMixin:
    """
    Auto creating serializer-based filter.
    """
    metadata_class = VueTableMetadata
    filter_backends = api_settings.DEFAULT_FILTER_BACKENDS

    def options(self, request, *args, **kwargs):
        """
        Handler method for HTTP 'OPTIONS' request.
        """
        metadata_class = getattr(self, "metadata_class", None)
        if metadata_class is None:
            return getattr(self, "http_method_not_allowed")(request, *args, **kwargs)

        data = metadata_class().determine_metadata(request, self)
        return Response(data)


class ExtGenericViewSet(GenericViewSet):
    _default_key = "default"
    queryset_function_name = "process_queryset"
    # The filter backend classes to use for queryset filtering

    def get_serializer_class(self):
        """
        Return the class to use for the serializer.
        Defaults to using `self.serializer_class`.

        You may want to override this if you need to provide different
        serializations depending on the incoming request.

        (Eg. admins get full serialization, others get basic serialization)
        """
        assert self.serializer_class is not None, (
                "'%s' should either include a `serializer_class` attribute, "
                "or override the `get_serializer_class()` method." % self.__class__.__name__
        )
        if isinstance(self.serializer_class, dict):  # 多个serializer_class
            assert (
                    self._default_key in self.serializer_class
            ), f"多个serializer时serializer_class必须包含下列key:{self._default_key}"
            if self.serializer_class.get(self.action):
                return self.serializer_class.get(self.action)
            else:
                return self.serializer_class.get(self._default_key)

        return self.serializer_class

    def data_permissions(self, request, view, queryset):
        """
        检查数据权限
        """
        for permission in self.get_permissions():
            if hasattr(permission, "data_permission"):
                return permission.data_permission(request, view, queryset)

    def get_queryset(self):
        """
        Get the list of items for this view.
        This must be an iterable, and may be a queryset.
        Defaults to using `self.queryset`.

        This method should always be used rather than accessing `self.queryset`
        directly, as `self.queryset` gets evaluated only once, and those results
        are cached for all subsequent requests.

        You may want to override this if you need to provide different
        querysets depending on the incoming request.

        (Eg. return a list of items that is specific to the user)
        """
        assert self.queryset is not None, (
                "'%s' should either include a `queryset` attribute, "
                "or override the `get_queryset()` method."
                % self.__class__.__name__
        )

        queryset = self.queryset
        if isinstance(queryset, QuerySet):
            # Ensure queryset is re-evaluated on each request.
            queryset = queryset.all()
            # Perform optimization on queryset
            serializer_class = self.get_serializer_class()
            if hasattr(serializer_class, self.queryset_function_name):
                queryset = getattr(serializer_class, self.queryset_function_name)(self.request, queryset)

        return queryset


class ExportMixin:
    """
    Export data to csv/xlsx file
    """
    renderer_classes = api_settings.DEFAULT_RENDERER_CLASSES + [CustomCSVRenderer, CustomExcelRenderer]

    def get_export_columns(self):
        """
        获取导出列信息
        :return:
        """
        serializer = self.get_serializer()  # noqa
        fields = serializer._readable_fields
        columns = {}
        for field in fields:
            field_name = field.field_name
            if not field.style.get("column_visible", True):
                continue

            columns[field_name] = {"column_name": str(field.label)}
            if not isinstance(field, (ManyRelatedField, RelatedField)) and getattr(field, "choices", None):
                columns[field_name]["choices"] = dict(field.choices)
        return columns

    def get_renderer_context(self):
        context = super().get_renderer_context()  # noqa
        export_columns = self.get_export_columns()
        context['header'] = (
            self.request.GET['fields'].split(',')  # noqa
            if 'fields' in self.request.GET else export_columns.keys())  # noqa
        context['labels'] = {
            field_name: attrs["column_name"] for field_name, attrs in export_columns.items()
        }
        context['value_mapping'] = {
            field_name: attrs["choices"] for field_name, attrs in export_columns.items() if "choices" in attrs
        }
        return context



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['drfexts', 'drfexts.filtersets', 'drfexts.serializers']

package_data = \
{'': ['*']}

install_requires = \
['char>=0.1.2,<0.2.0',
 'conditions>=0.2.0,<0.3.0',
 'django-currentuser>=0.5.3,<0.6.0',
 'django-filter>=21.1',
 'django-storages>=1.12.3',
 'django>=3.2,<4.0',
 'djangorestframework-csv>=2.1.1,<3.0.0',
 'djangorestframework>=3.12.4',
 'openpyxl>=3.0.9',
 'orjson>=3.6.7,<4.0.0',
 'pip>=21.3.1',
 'pydantic==1.8.2']

setup_kwargs = {
    'name': 'drfexts',
    'version': '0.10.5',
    'description': 'Django Restframework Utils',
    'long_description': "drfexts\n=======\n\n[![GitHub license](https://img.shields.io/github/license/aiden520/drfexts)](https://github.com/aiden520/drfexts/blob/master/LICENSE)\n[![pypi-version](https://img.shields.io/pypi/v/drfexts.svg)](https://pypi.python.org/pypi/drfexts)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/drfexts)\n[![PyPI - Django Version](https://img.shields.io/badge/django-%3E%3D3.0-44B78B)](https://www.djangoproject.com/)\n[![PyPI - DRF Version](https://img.shields.io/badge/djangorestframework-%3E%3D3.0-red)](https://www.django-rest-framework.org)\n[![Build Status](https://app.travis-ci.com/aiden520/drfexts.svg?branch=master)](https://app.travis-ci.com/aiden520/drfexts)\n\n**Extensions for Django REST Framework**\n\n**Author:** aiden lu\n\nInstallation\n------------\n\n``` {.bash}\n$ pip install drfexts\n```\n\nUsage\n-----\n\n*views.py*\n\n``` {.python}\nfrom rest_framework.views import APIView\nfrom rest_framework.settings import api_settings\nfrom drfexts.viewsets import ExportMixin\n\nclass MyView (ExportMixin, APIView):\n    ...\n```\n\nOrdered Fields\n--------------\n\nBy default, a `CSVRenderer` will output fields in sorted order. To\nspecify an alternative field ordering you can override the `header`\nattribute. There are two ways to do this:\n\n1)  Create a new renderer class and override the `header` attribute\n    directly:\n\n    > ``` {.python}\n    > class MyUserRenderer (CSVRenderer):\n    >     header = ['first', 'last', 'email']\n    >\n    > @api_view(['GET'])\n    > @renderer_classes((MyUserRenderer,))\n    > def my_view(request):\n    >     users = User.objects.filter(active=True)\n    >     content = [{'first': user.first_name,\n    >                 'last': user.last_name,\n    >                 'email': user.email}\n    >                for user in users]\n    >     return Response(content)\n    > ```\n\n2)  Use the `renderer_context` to override the field ordering on the\n    fly:\n\n    > ``` {.python}\n    > class MyView (APIView):\n    >     renderer_classes = [CSVRenderer]\n    >\n    >     def get_renderer_context(self):\n    >         context = super().get_renderer_context()\n    >         context['header'] = (\n    >             self.request.GET['fields'].split(',')\n    >             if 'fields' in self.request.GET else None)\n    >         return context\n    >\n    >     ...\n    > ```\n\nLabeled Fields\n--------------\n\nCustom labels can be applied to the `CSVRenderer` using the `labels`\ndict attribute where each key corresponds to the header and the value\ncorresponds to the custom label for that header.\n\n1\\) Create a new renderer class and override the `header` and `labels`\nattribute directly:\n\n> ``` {.python}\n> class MyBazRenderer (CSVRenderer):\n>     header = ['foo.bar']\n>     labels = {\n>         'foo.bar': 'baz'\n>     }\n> ```\n\nPagination\n----------\n\nUsing the renderer with paginated data is also possible with the new\n[PaginatedCSVRenderer]{.title-ref} class and should be used with views\nthat paginate data\n\nFor more information about using renderers with Django REST Framework,\nsee the [API\nGuide](http://django-rest-framework.org/api-guide/renderers/) or the\n[Tutorial](http://django-rest-framework.org/tutorial/1-serialization/).\n\nRunning the tests\n-----------------\n\nTo run the tests against the current environment:\n\n``` {.bash}\n$ ./manage.py test\n```\n\n### Changelog\n\n1.0.0\n-----\n\n-   Initial release\n\n## Thanks\n\n[![PyCharm](docs/pycharm.svg)](https://www.jetbrains.com/?from=drfexts)",
    'author': 'aiden',
    'author_email': 'allaher@icloud.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aiden520/drfexts',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



import warnings

from django.contrib.postgres.search import SearchVector, SearchQuery
from django.core.exceptions import ImproperlyConfigured
from django_filters.rest_framework import DjangoFilterBackend, filterset
from django_filters.filters import (
    Filter,
    BooleanFilter,
    CharFilter,
    TimeFilter,
    UUIDFilter,
)
from django_filters.utils import get_model_field
from rest_framework import serializers
from rest_framework.utils.field_mapping import ClassLookupDict
from rest_framework.filters import OrderingFilter, SearchFilter
from .filters import (
    IsNullFilter,
    IsNotNullFilter,
    ExtendedMultipleChoiceFilter,
    ExtendedNumberFilter,
    MultipleSelectFilter,
    ExtendedDateFromToRangeFilter,
    ExtendedModelMultipleChoiceFilter,
    ExtendedCharFilter,
    ExtendedDisplayMultipleChoiceFilter,
)
from ..serializers.fields import IsNotNullField, IsNullField, DisplayChoiceField

BOOLEAN_CHOICES = (
    ('false', 'False'),
    ('true', 'True'),
)

FILTER_FOR_SERIALIZER_FIELD_DEFAULTS = ClassLookupDict(
    {
        serializers.IntegerField: {'filter_class': ExtendedNumberFilter},
        serializers.DateField: {'filter_class': ExtendedDateFromToRangeFilter},
        serializers.DateTimeField: {'filter_class': ExtendedDateFromToRangeFilter},
        serializers.TimeField: {'filter_class': TimeFilter},
        serializers.DecimalField: {'filter_class': ExtendedNumberFilter},
        serializers.FloatField: {'filter_class': ExtendedNumberFilter},
        serializers.BooleanField: {'filter_class': BooleanFilter},
        serializers.EmailField: {'filter_class': CharFilter},
        serializers.FileField: {'filter_class': CharFilter},
        serializers.URLField: {'filter_class': CharFilter},
        serializers.UUIDField: {'filter_class': UUIDFilter},
        serializers.PrimaryKeyRelatedField: {
            'filter_class': ExtendedModelMultipleChoiceFilter,
            'extra': lambda f: {"queryset": f.queryset, "distinct": False},
        },
        serializers.SlugRelatedField: {
            'filter_class': ExtendedModelMultipleChoiceFilter,
            'extra': lambda f: {"queryset": f.queryset, "to_field_name": f.slug_field, "distinct": False},
        },
        serializers.ManyRelatedField: {
            'filter_class': ExtendedModelMultipleChoiceFilter,
            'extra': lambda f: {"queryset": f.child_relation.queryset},
        },
        serializers.RelatedField: {'filter_class': MultipleSelectFilter},
        serializers.JSONField: {'filter_class': CharFilter, 'extra': lambda f: {"lookup_expr": "icontains"}},
        serializers.ListField: {'filter_class': CharFilter, 'extra': lambda f: {"lookup_expr": "contains"}},
        IsNotNullField: {'filter_class': IsNotNullFilter},
        IsNullField: {'filter_class': IsNullFilter},
        serializers.ReadOnlyField: {'filter_class': CharFilter},
        DisplayChoiceField: {
            'filter_class': ExtendedDisplayMultipleChoiceFilter,
            'extra': lambda f: {"choices": list(f.choices.items()), 'distinct': False},
        },
        serializers.ChoiceField: {
            'filter_class': ExtendedMultipleChoiceFilter,
            'extra': lambda f: {"choices": list(f.choices.items()), "distinct": False},
        },
        serializers.CharField: {'filter_class': ExtendedCharFilter},
    }
)


class InitialFilterSet(filterset.FilterSet):
    def __init__(self, data=None, *args, **kwargs):
        # if filterset is bound, use initial values as defaults
        if data is not None:
            # get a mutable copy of the QueryDict
            data = data.copy()

            for name, f in self.base_filters.items():
                initial = f.extra.get('initial')

                # filter param is either missing or empty, use initial as default
                if not data.get(name) and initial:
                    data[name] = initial

        super().__init__(data, *args, **kwargs)


class AutoFilterBackend(DjangoFilterBackend):
    """
    Generate filterset class for
    """

    filterset_base = InitialFilterSet

    def get_filterset_class(self, view, queryset=None):
        """
        Return the `FilterSet` class used to filter the queryset.
        """
        filterset_class = getattr(view, 'filterset_class', None)
        filterset_fields_overwrite = getattr(view, 'filterset_fields_overwrite', {})

        # full text search

        if filterset_class:
            filterset_model = filterset_class._meta.model  # noqa

            # FilterSets do not need to specify a Meta class
            if filterset_model and queryset is not None:
                assert issubclass(
                    queryset.model, filterset_model
                ), 'FilterSet model %s does not match queryset model %s' % (filterset_model, queryset.model)

            return filterset_class

        serializer = view.get_serializer()

        if not isinstance(serializer, serializers.ModelSerializer):
            return None

        filterset_model = serializer.Meta.model  # noqa
        filterset_fields = {}

        overwrite_fields = {k: v for k, v in filterset_fields_overwrite.items() if isinstance(v, Filter)}
        overwrite_kwargs = {k: v for k, v in filterset_fields_overwrite.items() if isinstance(v, dict)}

        for filter_name, field in serializer.fields.items():
            if (
                getattr(field, "write_only", False)
                or field.source == "*"
                or isinstance(field, serializers.BaseSerializer)
            ):
                continue

            field_name = field.source.replace(".", "__") or filter_name
            if get_model_field(filterset_model, field_name) is None and (
                queryset is not None and filter_name not in queryset.query.annotations
            ):
                continue

            try:
                filter_spec = FILTER_FOR_SERIALIZER_FIELD_DEFAULTS[field]
            except KeyError:
                warnings.warn(f"{filter_name} 字段未找到过滤器, 跳过自动成filter!")
                continue

            extra = filter_spec.get("extra")
            kwargs = {"field_name": field_name, "label": field.label, "help_text": field.help_text}
            if callable(extra):
                kwargs.update(extra(field))

            if "queryset" in kwargs and kwargs["queryset"] is None:
                warnings.warn(f"{filter_name} 字段未提供queryset, 跳过自动成filter!")
                continue

            overwrite_value = overwrite_kwargs.get(filter_name)
            if overwrite_value:
                kwargs.update(overwrite_value)

            filterset_field = filter_spec["filter_class"](**kwargs)
            filterset_fields[filter_name] = filterset_field

        filterset_fields.update(overwrite_fields)

        AutoFilterSet = type("AutoFilterSet", (self.filterset_base,), filterset_fields)  # noqa
        return AutoFilterSet


class OrderingFilterBackend(OrderingFilter):
    """
    Extra supporting for ordering by serializer field name
    """

    def get_serializer_class(self, view):
        # If `ordering_fields` is not specified, then we determine a default
        # based on the serializer class, if one exists on the view.
        if hasattr(view, "get_serializer_class"):
            try:
                serializer_class = view.get_serializer_class()
            except AssertionError:
                # Raised by the default implementation if
                # no serializer_class was found
                serializer_class = None
        else:
            serializer_class = getattr(view, "serializer_class", None)

        if serializer_class is None:
            msg = (
                "Cannot use %s on a view which does not have either a "
                "'serializer_class', an overriding 'get_serializer_class' "
                "or 'ordering_fields' attribute."
            )
            raise ImproperlyConfigured(msg % self.__class__.__name__)

        return serializer_class

    def get_ordering(self, request, queryset, view):
        """
        Ordering is set by a comma delimited ?ordering=... query parameter.

        The `ordering` query parameter can be overridden by setting
        the `ordering_param` value on the OrderingFilter or by
        specifying an `ORDERING_PARAM` value in the API settings.
        """
        params = request.query_params.get(self.ordering_param)
        if params:
            fields = [param.strip() for param in params.split(",")]
            fields = self.get_fixed_fields(fields, view, request)
            ordering = self.remove_invalid_fields(queryset, fields, view, request)
            if ordering:
                return ordering

        # No ordering was included, or all the ordering fields were invalid
        return self.get_default_ordering(view)

    def get_default_valid_fields(self, queryset, view, context={}):
        serializer_class = self.get_serializer_class(view)
        return [
            (field.source.replace(".", "__") or field_name, field.label)
            for field_name, field in serializer_class(context=context).fields.items()
            if not getattr(field, "write_only", False) and not field.source == "*"
        ]

    def get_fixed_fields(self, fields, view, request):
        """
        Get query name by field name
        """
        serializer_class = self.get_serializer_class(view)
        default_fields = serializer_class(context={"request": request}).fields
        fixed_fields = []
        for field_name in fields:
            prefix = "-" if field_name.startswith("-") else ""
            field_name = field_name.lstrip("-")
            field = default_fields.get(field_name)
            if field and not getattr(field, "write_only", False) and not field.source == "*":
                fixed_name = field.source.replace(".", "__")
                fixed_fields.append(prefix + (fixed_name or field_name))
            else:
                fixed_fields.append(prefix + field_name)

        return fixed_fields


class FullTextSearchFilter(SearchFilter):
    """
    Search filter that supports fulltext search
    """
    search_vector = None
    search_query = None

    def get_search_vector(self):
        if isinstance(self.search_vector, (list, tuple)):
            return SearchVector(*self.search_vector)
        elif isinstance(self.search_vector, SearchVector):
            return self.search_vector
        elif self.search_vector is None:
            return

        raise ImproperlyConfigured("`search_vector` must be type of list, tuple or 'SearchVector' instance.")

    def get_search_query(self, search_terms):
        if self.search_query is None:
            return SearchQuery(' '.join(search_terms))
        elif isinstance(self.search_query, SearchQuery):
            return self.search_query

        raise ImproperlyConfigured("`search_query` must be instance of 'SearchQuery'")

    def filter_queryset(self, request, queryset, view):
        search_terms = self.get_search_terms(request)
        if not search_terms:
            return queryset

        search_vector = self.get_search_vector()
        search_query = self.get_search_query(search_terms)

        if not search_vector or not search_query:
            return queryset

        return queryset.annotate(search=search_vector).filter(search=search_query)



import uuid
from functools import partial

from django.conf import settings
from django.contrib.postgres.fields import ArrayField as PGArrayField
from django.contrib.contenttypes import fields as ct_fields
from django.db import models
from django.core import checks
from django.db.models import CASCADE
from django_currentuser.db.models import CurrentUserField

from .constants import CommonStatus, AuditStatus
from .utils import get_serial_code


class DefaultHelpTextMixin:
    def __init__(self, verbose_name, *args, **kwargs):
        kwargs.setdefault('help_text', verbose_name)
        super().__init__(verbose_name, *args, **kwargs)


class NullHelpTextMixin:
    def __init__(self, verbose_name, *args, **kwargs):
        kwargs['null'] = False
        kwargs.setdefault('help_text', verbose_name)
        super().__init__(verbose_name, *args, **kwargs)


class RelatedNameCheckMixin:
    def check(self, **kwargs):
        return [
            *super().check(**kwargs),  # noqa
            *self._check_related_name(),
        ]

    def _check_related_name(self, **kwargs):
        return (
            [
                checks.Warning(
                    "Setting 'related_name' on a RelatedField may be better!",
                    obj=self,
                )
            ]
            if self.remote_field.related_name is None
            else []
        )  # noqa


class AutoField(models.AutoField):
    def __init__(self, verbose_name="主键", **kwargs):
        kwargs.setdefault('help_text', verbose_name)
        super().__init__(verbose_name, **kwargs)


class BigAutoField(DefaultHelpTextMixin, models.BigAutoField):
    pass


class CharField(NullHelpTextMixin, models.CharField):
    pass


class TextField(NullHelpTextMixin, models.TextField):
    pass


class IntegerField(DefaultHelpTextMixin, models.IntegerField):
    pass


class BigIntegerField(DefaultHelpTextMixin, models.BigIntegerField):
    pass


class SmallIntegerField(DefaultHelpTextMixin, models.SmallIntegerField):
    pass


class PositiveSmallIntegerField(DefaultHelpTextMixin, models.PositiveSmallIntegerField):
    pass


class BooleanField(DefaultHelpTextMixin, models.BooleanField):
    pass


class FileField(DefaultHelpTextMixin, models.FileField):
    pass


class ImageField(DefaultHelpTextMixin, models.ImageField):
    pass


class FilePathField(DefaultHelpTextMixin, models.FilePathField):
    pass


class FloatField(DefaultHelpTextMixin, models.FloatField):
    pass


class DecimalField(DefaultHelpTextMixin, models.DecimalField):
    pass


class DateTimeField(DefaultHelpTextMixin, models.DateTimeField):
    pass


class DateField(DefaultHelpTextMixin, models.DateField):
    pass


class TimeField(DefaultHelpTextMixin, models.TimeField):
    pass


class DurationField(DefaultHelpTextMixin, models.DurationField):
    pass


class EmailField(DefaultHelpTextMixin, models.EmailField):
    pass


class URLField(DefaultHelpTextMixin, models.URLField):
    pass


class IPAddressField(DefaultHelpTextMixin, models.IPAddressField):
    pass


class UUIDField(DefaultHelpTextMixin, models.UUIDField):
    pass


class JSONField(DefaultHelpTextMixin, models.JSONField):
    pass


class ArrayField(PGArrayField):
    def __init__(self, verbose_name, base_field, **kwargs):
        kwargs.setdefault('help_text', verbose_name)
        kwargs.setdefault('verbose_name', verbose_name)
        super().__init__(base_field, **kwargs)


class AutoUUIDField(models.UUIDField):
    def __init__(self, verbose_name="主键", **kwargs):
        kwargs['blank'] = True
        kwargs["default"] = uuid.uuid4
        kwargs.setdefault('help_text', verbose_name)
        kwargs.setdefault("primary_key", True)
        super().__init__(verbose_name, **kwargs)


class DefaultCodeField(models.CharField):
    """
    自动编号字段
    """

    DEFAULT_LENGTH = 15

    def __init__(self, verbose_name="编号", prefix="", **kwargs):
        self.prefix = prefix
        kwargs['blank'] = True
        kwargs["default"] = partial(get_serial_code, prefix)
        kwargs["max_length"] = self.DEFAULT_LENGTH + len(prefix)
        kwargs['editable'] = False
        kwargs.setdefault("help_text", verbose_name)
        super().__init__(verbose_name, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        # Only include kwarg if it's not the default
        if self.prefix != "":
            kwargs['prefix'] = self.prefix
        return name, path, args, kwargs


class DescriptionField(models.TextField):
    """
    description = DescriptionField()
    """

    def __init__(self, verbose_name="描述", **kwargs):
        kwargs.setdefault('blank', True)
        kwargs.setdefault('help_text', verbose_name)
        super().__init__(verbose_name, **kwargs)


class UserForeignKeyField(models.ForeignKey):
    """
    user = UserForeignKeyField()
    """

    def __init__(self, verbose_name="关联的用户", to=None, on_delete=None, **kwargs):
        to = to or settings.AUTH_USER_MODEL
        on_delete = on_delete or CASCADE
        kwargs.setdefault("db_constraint", False)
        kwargs.setdefault('help_text', verbose_name)
        super().__init__(to=to, verbose_name=verbose_name, on_delete=on_delete, **kwargs)


class UpdatedAtField(models.DateTimeField):
    """
    update_datetime = ModifyDateTimeField()
    """

    def __init__(self, verbose_name="修改时间", **kwargs):
        kwargs['editable'] = False
        kwargs['auto_now'] = True
        kwargs.setdefault('help_text', '该记录的最后修改时间')
        kwargs.setdefault('blank', True)
        super().__init__(verbose_name, **kwargs)


class CreatedAtField(models.DateTimeField):
    """
    create_datetime = CreateDateTimeField()
    """

    def __init__(self, verbose_name="创建时间", **kwargs):
        kwargs['editable'] = False
        kwargs['auto_now_add'] = True
        kwargs.setdefault('help_text', '该记录的创建时间')
        kwargs.setdefault('blank', True)
        super().__init__(verbose_name, **kwargs)


class CreatedByField(CurrentUserField):
    """
    created_by = CreatedByField()
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('verbose_name', '创建人')
        kwargs.setdefault('help_text', '该记录的创建者')
        kwargs.setdefault('related_name', '%(class)s_created_by')
        kwargs.setdefault("on_delete", models.CASCADE)
        kwargs['db_constraint'] = False
        super().__init__(*args, **kwargs)

    def _warn_for_shadowing_args(self, *args, **kwargs):
        pass


class UpdatedByField(CurrentUserField):
    """
    updated_by = UpdatedByField()
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('on_update', True)
        kwargs.setdefault('verbose_name', '修改人')
        kwargs.setdefault('help_text', '该记录的修改人')
        kwargs.setdefault('related_name', '%(class)s_updated_by')
        kwargs.setdefault("on_delete", models.CASCADE)
        kwargs['db_constraint'] = False
        super().__init__(*args, **kwargs)

    def _warn_for_shadowing_args(self, *args, **kwargs):
        pass


class CreatorCharField(models.CharField):
    """
    creator = CreatorCharField()
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 128)
        kwargs.setdefault('null', True)
        kwargs.setdefault('blank', True)
        kwargs.setdefault('verbose_name', '创建者')
        kwargs.setdefault('help_text', '该记录的创建者')
        super().__init__(*args, **kwargs)


class ModifierCharField(models.CharField):
    """
    modifier = ModifierCharField()
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 128)
        kwargs.setdefault('null', True)
        kwargs.setdefault('blank', True)
        kwargs.setdefault('verbose_name', '修改者')
        kwargs.setdefault('help_text', '该记录最后修改者')
        super().__init__(*args, **kwargs)


class StatusField(models.PositiveSmallIntegerField):
    """
    status = StatusField()
    """

    def __init__(self, verbose_name="状态", **kwargs):
        kwargs.setdefault('choices', CommonStatus.choices)
        kwargs.setdefault('default', CommonStatus.VALID)
        kwargs.setdefault('help_text', '100：已失效，75：待失效，50：有效，25：暂停中，10：待生效，5：待提交，0：删除')
        super().__init__(verbose_name, **kwargs)


class AuditStatusField(models.PositiveSmallIntegerField):
    """
    status = StatusField()
    """

    def __init__(self, verbose_name="审核状态", **kwargs):
        kwargs.setdefault('choices', AuditStatus.choices)
        kwargs.setdefault('null', True)
        kwargs.setdefault('blank', True)
        kwargs.setdefault('help_text', '该记录的审核状态')
        super().__init__(verbose_name, **kwargs)


class VirtualForeignKey(models.ForeignKey):
    def __init__(self, verbose_name, to, *args, **kwargs):
        kwargs.setdefault("verbose_name", verbose_name)
        kwargs['db_constraint'] = False

        if kwargs.get("null"):
            kwargs.setdefault("blank", True)
            kwargs.setdefault("on_delete", models.SET_NULL)
        else:
            kwargs.setdefault("on_delete", models.CASCADE)

        super().__init__(to, *args, **kwargs)


class OneToOneField(models.OneToOneField):
    def __init__(self, verbose_name, to, *args, **kwargs):
        kwargs.setdefault("verbose_name", verbose_name)
        kwargs["db_constraint"] = False

        if kwargs.get("null"):
            kwargs.setdefault("blank", True)
            kwargs.setdefault("on_delete", models.SET_NULL)
        else:
            kwargs.setdefault("on_delete", models.CASCADE)

        super().__init__(to, *args, **kwargs)


class VirtualManyToMany(models.ManyToManyField):
    def __init__(self, verbose_name, to, *args, **kwargs):
        kwargs.setdefault("verbose_name", verbose_name)
        if "through" not in kwargs:
            kwargs["db_constraint"] = False

        super().__init__(to, *args, **kwargs)


class GenericForeignKey(ct_fields.GenericForeignKey):
    def __init__(self, ct_field='content_type', fk_field='object_id', for_concrete_model=True):
        super().__init__(ct_field=ct_field, fk_field=fk_field, for_concrete_model=for_concrete_model)


class GenericRelation(ct_fields.GenericRelation):
    def __init__(
        self,
        verbose_name,
        to,
        object_id_field='object_id',
        content_type_field='content_type',
        for_concrete_model=True,
        related_query_name=None,
        limit_choices_to=None,
        **kwargs,
    ):
        super().__init__(
            to,
            object_id_field=object_id_field,
            content_type_field=content_type_field,
            for_concrete_model=for_concrete_model,
            related_query_name=related_query_name,
            limit_choices_to=limit_choices_to,
            verbose_name=verbose_name,
            **kwargs,
        )



import copy
from enum import Enum
from typing import Optional, List, Union, Any, Dict
from pydantic import BaseModel


class TypeEnum(str, Enum):
    """
    列类型
    """
    seq = 'seq'
    checkbox = 'checkbox'
    radio = 'radio'
    expand = 'expand'
    html = 'html'


class FixedEnum(str, Enum):
    """
    列固定
    """
    left = 'left'  # 固定左侧
    right = 'right'  # 固定右侧


class AlignEnum(str, Enum):
    """
    对齐方式
    """
    left = 'left'
    center = 'center'
    right = 'right'


class OverflowEnum(str, Enum):
    """
    内容过长时展示行为
    """
    ellipsis = 'ellipsis'  # 只显示省略号
    title = 'title'  # 并且显示为原生
    tooltip = 'tooltip'  # 并且显示为提示


class FieldTypeEnum(str, Enum):
    """
    字段类型
    """
    auto = 'auto'
    number = 'number'
    string = 'string'


class Filter(BaseModel):
    """
    过滤器配置项
    """
    label: str  # 显示的值
    value: Any  # 实际的值
    checked: bool = False  # 默认是否选中
    resetValue: Optional[Any]  # 重置时的默认值
    data: Optional[Any]  # 自定义渲染的数据值（当使用自定义模板时可能会用到）


class CellRender(BaseModel):
    """
    默认的渲染器配置项
    """
    name: str  # 渲染器名称
    props: Any  # 渲染的参数（请查看目标渲染的 Props）
    options: List[Any]  # 只对 name=select 有效，下拉选项列表
    optionProps: Optional[Any]  # 只对 name=select 有效，下拉选项属性参数配置
    optionGroups: List[Any]  # 只对 name=select 有效，下拉分组选项列表
    optionGroupProps: Optional[Any]  # 只对 name=select 有效，下拉分组选项属性参数配置
    events: Optional[Any]
    content: Optional[str]


class EditRender(BaseModel):
    """
    可编辑渲染器配置项
    """
    name: str  # 渲染器名称
    enabled: bool = True  # 是否启用
    props: Optional[Any]  # 渲染的参数（请查看目标渲染的 Props）
    options: Optional[List]  # 只对 name=select 有效，下拉选项列表
    optionProps: Optional[Any]  # 只对 name=select 有效，下拉选项属性参数配置
    optionGroups: Optional[List]  # 只对 name=select 有效，下拉分组选项列表
    optionGroupProps: Optional[Any]  # 只对 name=select 有效，下拉分组选项属性参数配置
    events: Optional[Any]
    content: Optional[str]  # 渲染组件的内容（仅用于特殊组件）
    autofocus: Optional[str]  # 如果是自定义渲染可以指定聚焦的选择器，例如 .my-input
    autoselect: bool = False  # 是否在激活编辑之后自动选中输入框内容
    defaultValue: Optional[Any] = None
    immediate: bool = False  # 输入值实时同步更新（默认情况下单元格编辑的值只会在被触发时同步，如果需要实时同步可以设置为 true）
    placeholder: Optional[str]  # 单元格占位符，但单元格为空值时显示的占位符


class ContentRender(BaseModel):
    """
    内容渲染配置项
    """
    name: str  # 渲染器名称
    props: Optional[Any]  # 渲染的参数（请查看目标渲染的 Props）
    events: Optional[Any]


class Style(BaseModel):
    """
    Properties of Vue-Table-Column.
    """
    type: Optional[TypeEnum]  # 列的类型
    field: Optional[str]  # 列字段名（注：属性层级越深，渲染性能就越差，例如：aa.bb.cc.dd.ee）
    title: Optional[str]  # 列标题（支持开启国际化）
    width: Union[int, str, None]  # 列宽度（如果为空则均匀分配剩余宽度，如果全部列固定了，可能会存在宽屏下不会铺满，可以配合 &quot;%&quot; 或者 &quot;min-width&quot; 布局）
    minWidth: Union[int, str, None]  # 最小列宽度；会自动将剩余空间按比例分配
    resizable: Optional[bool]  # 列是否允许拖动列宽调整大小
    visible: bool = True  # 默认是否显示
    fixed: Optional[FixedEnum]  # 将列固定在左侧或者右侧（注意：固定列应该放在左右两侧的位置）
    align: Optional[AlignEnum]  # 列对齐方式
    headerAlign: Optional[AlignEnum]  # 表头列对齐方式
    footerAlign: Optional[AlignEnum]  # 表尾列对齐方式
    showOverflow: Optional[OverflowEnum]  # 内容过长时显示为省略号
    showHeaderOverflow: Optional[OverflowEnum]  # 当表头内容过长时显示为省略号
    showFooterOverflow: Optional[OverflowEnum]  # 当表尾内容过长时显示为省略号
    className: Optional[str]  # 给单元格附加 className
    headerClassName: Optional[str]  # 给表头的单元格附加 className
    footerClassName: Optional[str]  # 给表尾的单元格附加 className
    formatter: Optional[Any]  # 格式化显示内容
    sortable: bool = False  # 是否允许排序
    sortBy: Optional[str]  # 只对 sortable 有效，指定排序的字段（当值 formatter 格式化后，可以设置该字段，使用值进行排序）
    sortType: FieldTypeEnum = FieldTypeEnum.auto  # 排序的字段类型，比如字符串转数值等
    filters: Optional[List[Filter]]  # 配置筛选条件（注：筛选只能用于列表，如果是树结构则过滤根节点）
    filterMultiple: bool = True  # 只对 filters 有效，筛选是否允许多选
    filterMethod: Optional[Any]  # 只对 filters 有效，列的筛选方法，该方法的返回值用来决定该行是否显示
    cellType: FieldTypeEnum = FieldTypeEnum.auto  # 只对特定功能有效，单元格值类型（例如：导出数据类型设置）
    cellRender: Optional[CellRender]  # 默认的渲染器配置项
    editRender: Optional[EditRender]  # 可编辑渲染器配置项
    contentRender: Optional[ContentRender]  # 内容渲染器配置项
    treeNode: bool = False  # 只对 tree-config 配置时有效，指定为树节点
    params: Optional[Any]  # 额外的参数（可以用来存放一些私有参数）
    children: Optional[List]  # 表头分组






import os
from urllib.parse import urljoin

from django.conf import settings
from django.utils.encoding import force_text
from storages.backends.s3boto3 import S3Boto3Storage


class NormalizeNameMixin:
    def _normalize_name(self, name):
        # urljoin won't work if name is absolute path
        name = name.lstrip('/')
        base_path = force_text(self.location)
        final_path = urljoin(base_path + "/", name)
        name = os.path.normpath(final_path.lstrip('/'))

        # Add / to the end of path since os.path.normpath will remove it
        if final_path.endswith('/') and not name.endswith('/'):
            name += '/'

        # Store filenames with forward slashes, even on Windows.
        return name.replace('\\', '/')


class AliOSSMediaStorage(NormalizeNameMixin, S3Boto3Storage):
    location = settings.MEDIAFILES_LOCATION
    file_overwrite = False


class AliOSSStaticStorage(NormalizeNameMixin, S3Boto3Storage):
    location = settings.STATICFILES_LOCATION
    file_overwrite = True



"""
The metadata API is used to allow customization of how `OPTIONS` requests
are handled. We currently provide a single default implementation that returns
some fairly ad-hoc information about the view.
Future implementations might use JSON schema or other definitions in order
to return this information in a more standardized way.
"""
from rest_framework.fields import ChoiceField, MultipleChoiceField
from rest_framework.metadata import BaseMetadata
from django.core.exceptions import PermissionDenied
from django.http import Http404
import copy
from django.utils.encoding import force_str
from rest_framework import exceptions, serializers
from rest_framework.request import clone_request
from rest_framework.utils.field_mapping import ClassLookupDict
from .style import Style
from .utils import to_table_choices

SIMPLE = 'simple'
COMPLEX = 'subtable'


class TableMeta(dict):
    __slots__ = ('expand', 'table_type', 'fields', 'data_field')

    def __init__(self, table_type=SIMPLE, expand=False, **kwargs):
        self.expand = expand
        if table_type not in (SIMPLE, COMPLEX):
            raise ValueError('Unknown table type.')
        self.table_type = table_type
        _ = dict()
        for attr in self.__slots__:
            if getattr(self, attr, None):
                _.update({attr: getattr(self, attr)})
        super().__init__(_, **kwargs)


class VueTableMetadata(BaseMetadata):
    """
    This is the default metadata implementation.
    It returns an ad-hoc set of information about the view.
    There are not any formalized standards for `OPTIONS` responses
    for us to base this on.
    """

    style_class = Style
    label_lookup = ClassLookupDict(
        {
            serializers.Field: 'field',
            serializers.BooleanField: 'boolean',
            serializers.NullBooleanField: 'boolean',
            serializers.CharField: 'string',
            serializers.UUIDField: 'string',
            serializers.URLField: 'url',
            serializers.EmailField: 'email',
            serializers.RegexField: 'regex',
            serializers.SlugField: 'slug',
            serializers.IntegerField: 'integer',
            serializers.FloatField: 'float',
            serializers.DecimalField: 'decimal',
            serializers.DateField: 'date',
            serializers.DateTimeField: 'datetime',
            serializers.TimeField: 'time',
            serializers.ChoiceField: 'choice',
            serializers.MultipleChoiceField: 'multiple choice',
            serializers.FileField: 'file upload',
            serializers.ImageField: 'image upload',
            serializers.ListField: 'list',
            serializers.DictField: 'nested object',
            serializers.Serializer: 'nested object',
        }
    )

    search_type_lookup = ClassLookupDict(
        {
            serializers.Field: 'input',
            serializers.BooleanField: 'select',
            serializers.NullBooleanField: 'select',
            serializers.IntegerField: 'num',
            serializers.FloatField: 'input',
            serializers.DecimalField: 'input',
            serializers.DateField: 'date',
            serializers.DateTimeField: 'date',
            serializers.TimeField: 'date',
            serializers.ChoiceField: 'select',
            serializers.MultipleChoiceField: 'select',
        }
    )

    def determine_metadata(self, request, view):
        metadata = dict()
        metadata['name'] = view.get_view_name()
        metadata['description'] = view.get_view_description()
        metadata['renders'] = [renderer.media_type for renderer in view.renderer_classes]
        metadata['parses'] = [parser.media_type for parser in view.parser_classes]
        if hasattr(view, 'get_serializer'):
            actions = self.determine_actions(request, view)
            if actions:
                metadata['actions'] = actions
        return metadata

    def determine_actions(self, request, view):
        """
        For generic class based views we return information about
        the fields that are accepted for 'PUT' and 'POST' methods.
        """
        setattr(self, "request", request)
        setattr(self, "view", view)
        actions = {}
        action_names = []
        extra_actions = view.get_extra_actions()
        extra_action_names = [
            act.mapping['get']
            for act in extra_actions
            if 'get' in act.mapping and act.mapping['get'] not in ("export", "meta")
        ]
        if 'get' in view.action_map:
            action_names = [view.action_map["get"]] + extra_action_names

        for action in action_names:
            actions[action] = dict()
            origin_action = view.action
            view.action = action
            serializer = view.get_serializer()
            actions[action]["meta"] = self.get_serializer_meta(view, serializer)
            actions[action]["columns"] = self.get_serializer_info(serializer)
            view.action = origin_action

        # TODO check permissions
        # for method in {'GET'} & set(view.allowed_methods):
        #     view.request = clone_request(request, method)
        #     try:
        #         # Test global permissions
        #         if hasattr(view, 'check_permissions'):
        #             view.check_permissions(view.request)
        #     except (exceptions.APIException, PermissionDenied, Http404):
        #         pass
        #     else:
        #         # If user has appropriate permissions for the view, include
        #         # appropriate metadata about the fields that should be supplied.
        #         serializer = view.get_serializer()
        #         actions[method] = self.get_serializer_info(serializer)
        #     finally:
        #         view.request = request

        return actions

    def get_serializer_meta(self, view, serializer):
        """
        :param view: [description]
        :type view: [type]
        :param serializer: [description]
        :type serializer: [type]
        """
        table_meta = getattr(view, "table_meta", {})
        data_field_name = table_meta.get('data_field')
        data_field = serializer.fields.get(data_field_name)
        if data_field and hasattr(data_field, "child"):
            table_meta['fields'] = self.get_serializer_info(data_field.child)

        return table_meta

    def get_serializer_info(self, serializer):
        """
        Given an instance of a serializer, return a dictionary of metadata
        about its fields.
        """
        if hasattr(serializer, 'child'):
            # If this is a `ListSerializer` then we want to examine the
            # underlying child serializer instance instead.
            serializer = serializer.child

        return [
            self.get_field_style(field)
            for field_name, field in serializer.fields.items()
            if not isinstance(field, serializers.HiddenField) and not field.write_only
        ]

    def get_field_style(self, field, sub_table=True):
        """
        Given an instance of a serializer field, return a dictionary
        of metadata about it.
        """
        # validate style
        if isinstance(field.style, dict):
            copy_style = copy.deepcopy(field.style)
            style = self.style_class(**copy_style)
        elif isinstance(field.style, self.style_class):
            style = field.style
        else:
            raise ValueError("Unsupported style type.")

        style.field = field.field_name
        style.title = field.label

        if isinstance(field, (ChoiceField, MultipleChoiceField)) and field.choices:
            style.cellRender = to_table_choices(field.choices)

        return style.dict(exclude_none=True)
